;+
;This routine returns a structure or an array of structures
;read from the file provided.
;
;Only 2 arguments are needed if (a) filename is an index file or 
;(b) all the records are needed
;if keyword quant is set, then rec (or recnum) contains just the
;data from the structure element with tag-name quant
;
;In most cases, this routine is intelligent enough to choose the
;correct structure (using the subroutine getstructinfo.pro).  It
;determines the type of the file and the version based on the
;filename.  The version can be overridden with the vers keyword
;(ie, for experimental files).
;
;Four arguments should be used when the intention is to make a 
;series of calls to this routine using the same datafile with 
;one record being output each time.  The final argument should
;be undefined on the first call and then should not be changed
;for subsequent calls (it is the file unit number).  The third
;argument (rec) also should not be changed between calls.
;After the final call, close the file and free the file unit number
;using close,inlun & free_lun,inlun
;
;$Header: /home/users/burton/sageprof-rcs/readstructs.pro,v 1.38 2003/08/21 15:47:51 burton Exp $
;-
pro readstructs,filename,recnum,rec,inlun,vers=vers,quant=quant,$
                error_return=error_return,$
                type_return=type_return,exten_return=exten_return
if n_params() lt 2 then begin
    print,'usage: readstructs,filename,rec (for index structure) or '
    print,'       readstructs,filename,recnum,rec (for others)'
endif
;if this is windows, the case shouldn't matter...
if (!version.os eq 'Win32') then filename=strupcase(filename)

typ = 'UNKNOWN'
;inlun is included so that a series of calls can be made
if n_params() eq 4 then begin
;check that inlun is defined (if not then it is output only)
    if keyword_set(inlun) then begin
        result = fstat(inlun)
        if result.name eq filename then begin
            typ = 'OLD'
            struct = rec(0)
        endif else begin
            close,inlun & free_lun,inlun
        endelse
    endif
endif
if typ ne 'OLD' or n_params() lt 4 then begin
;parse the filename to find out the structure type
    emanelif = string(reverse(byte(filename)))
;reverse search for the beginning of the instrument name
    p = strpos(emanelif,'I_EGAS')
;convert p to the index (of 'I') wrt the forward string
    p = strlen(filename)-p-1
;search for the beginning of the file type 
    q = strpos(filename,'_',p)+1
;search for the underscore that marks the end of the file type
    r = strpos(filename,'_',q)
;parse the instrument and type names
    inst = 'SAGE_I'+strmid(filename,p+1,q-p-1)
    typ = strmid(filename,q,r-q)
;extract the version from the filename
    if not keyword_set(vers) then begin
        substring = string(reverse(byte(inst+typ+'_')))
        vers = strlen(filename)-strpos(emanelif,substring)+7
        vers = strmid(filename,vers,4)
    endif
    getstructinfo,struct,typ,vers
endif

;if it's an index file, will want the first (and only) record
if (typ eq 'INDEX') then recnum = 0
;get the structure item number if quant was specified
if keyword_set(quant) then begin
    quant = strupcase(quant)
    tags = tag_names(struct)
    ix = where(tags eq quant)
    if ix(0) eq -1 then begin
        print,quant,': Tag name not recognized for this structure.'
        return
    endif else begin
        qnum = ix(0)
    endelse
endif

;prepare to read the file
error_return = 0
if (typ eq 'INDEX' or n_params() gt 2) then begin
    if typ ne 'OLD' then begin
        openr,inlun,filename,/get_lun,error=err
        if (err ne 0) then begin
            error_return = err
            print,!ERR_STRING
            return
        endif
    endif
;extract the desired record
    data = assoc(inlun,struct)
    on_ioerror,RETRY1
    tries = 0
RETRY1:
    if tries gt 0 then print,tries,' retrying to read ',filename,recnum
    tries = tries + 1
    rec = data(recnum)
    on_ioerror,NULL
;swap the data if it is not a DEC or Windows machine
    if (!version.os ne 'OSF' and !version.os ne 'Win32' and $
        !version.os ne 'linux') then begin
        rec = swap_endian(rec)
    endif
    if keyword_set(quant) then quant = rec.(qnum)
;close the file
    if (n_params() lt 4) then begin
        close,inlun & free_lun,inlun
    endif

;if it was an index file, the input arguments' names are offset
    if (typ eq 'INDEX') then recnum = rec
endif else if (n_params() eq 2) then begin
;in this case, the user wants all records
    openr,inlun,filename,/get_lun,error=err
    if (err ne 0) then begin
        error_return = err
        print,!ERR_STRING
        return
    endif
    data = assoc(inlun,struct)
    if keyword_set(quant) then begin
;figure out the type of array involved
        sz = size(struct.(qnum))
;add a dimension to sz to reflect the type of array desired
        sz = [sz,sz(sz(0)+2)]
        sz(sz(0)+2) = sz(sz(0)+1)
;make the added dimension have 930 elements
        sz(sz(0)+1) = 930
;increment the number of dimensions
        sz(0) = sz(0) + 1
;make the new array
        recnum = make_array(size=sz)
    endif else begin
        recnum = replicate(struct,930)
    endelse
    i = 0
    tries = 0
    on_ioerror,RETRY
    while not EOF(inlun) do begin
RETRY:
        if tries gt 0 then print,tries,' retrying to read ',filename,i 
        tries = tries+1
        rec = data(i)
        if (!version.os ne 'OSF' and !version.os ne 'Win32' and $
            !version.os ne 'linux') then begin
            rec = swap_endian(rec)
        endif
        if keyword_set(quant) then begin
;need to copy the array to recnum(:,:,:,i) but the exact number of
;dimensions is unknown.
            slice = 1 & for j=1,sz(0)-1 do slice = slice*sz(j)
            recnum(slice*i:slice*(i+1)-1) = rec.(qnum)
        endif else begin
            recnum(i) = rec
        endelse
        i = i + 1
        tries = 0
    endwhile
    on_ioerror,NULL
    if keyword_set(quant) then begin
        if sz(0) gt 1 then begin
            recnum = reform(recnum(0:slice*i-1),[sz(1:sz(0)-1),i])
        endif 
    endif else begin
        recnum = recnum(0:i-1)
    endelse
    close,inlun & free_lun,inlun
endif else begin
;what's going on in this case
    print,'usage: readstructs,filename,record_number,output_record'
    return
endelse

type_return = typ
if keyword_set(vers) then exten_return = vers
return
end
