pro subset_sage2_v620,LISTFILE=listfile

;  This procedure subsets SAGE II version 6.20 binary data files.
;  The subset may be specified in terms of a lat/lon box and/or
;  the parameters to be included.
;  The selected parameter values are written to an ASCII output file
;  along with certain file header information for profiles that
;  match the lat/lon criteria.
;
;  The procedure will operate either on one or more files selected
;  during execution or on a list of files contained in a file
;  with the full path name of the list file specified as a string
;  value of the LISTFILE keyword.  Note that the full path name need
;  not be specified if the listfile resides in the current IDL default
;  directory.
;
;  INPUTS:
;  No positional parameters
;  Keyword LISTFILE is used to specify the name (with full path if not
;  in the current default directory) of the file containing a list
;  of SAGE II species files to be processed.
;  If the LISTFILE keyword is not specified, a file selection dialog
;  box will be displayed and one or may SAGE II species files may be
;  selected for processing.
;  Dialog boxes are used to specify the lat/lon box bounds and the
;  parameters to be included in the subset.
;
;  OUTPUTS:
;  An ASCII (text) format output file is created using the name of the
;  species file with "_subset.dat"  apeended to it.  The output file
;  includes:
;  1)  subset information--
;      The name of the file that was subset.
;      The subset criteria used.
;  2)  File information from the index file corresponding to the
;      SAGE II species file
;  and, for each profile matching the subset criteria
;  3)  profile information from the index file
;  4)  the specified parameters from that profile
;
;  One output file is created for each SAGE II file selected for processing.
;  The output files are written to the directory containing the corresponding
;  SAGE II species file.
;
;  If no latitude or longitude criteria are matched, a message to that
;  effect is written to the output file.
;
;  Note that multiple subsets of the same input file will cause the output
;  file to be overwritten since it would have the same name, so rename
;  the output files between runs if the same SAGE II file is subset in more
;  than one way.
;
;  EXTERNAL ROUTINES USED:
;  All of these files should be in the same directory as this program
;  (subset_sage2_v620.pro):
;
;  get_latlon_box.pro - routine to create a GUI for specifying the
;    lat/lon box bounds of interest; the corresponding
;    event handler routine is get_latlon_box_event.pro
;  select_tool.pro - routine to create a selection list GUI for the
;    list of parameters to be included in the subset; the corresponding
;    event handler routine is select_tool_event.pro
;  Several additional routines from the standard SAGE II v6.20 read
;  software package are also used and are included with this package:
;       ephinfo_y2k.pro
;       getstructinfo.pro
;       indexinfo_60d.pro
;       indexinfo_610.pro
;       metinfo.pro
;       metinfo_610.pro
;       readstructs.pro
;       refractinfo)60d.pro
;       specinfo_600.pro
;       specinfo_610.pro
;       suninfo.pro
;       transinfo.pro
;
;  ADDITIONAL REQUIRED FILES:
;  sage2_params.txt - a list of the parameters that occur in SAGE II version
;    6.20 data files
;  This file must reside in the current IDL default directory.
;
;  ASSUMPTIONS:
;  SAGE II data files retain their original file names
;  The INDEX files corresponding to the SAGE II species files being used
;  reside in the same directory as the species files.
;  The sage2_params.txt file resides in the current IDL default
;  directory.
;  Note: This code was written to work with older versions of IDL, so
;  some parts of the code that could take advantage of newer IDL
;  functionality don't do so in order to remain compatible with the
;  older versions.
;
;  PROGRAM EXECUTION:
;  1) Start the idlde--on PC, click on the IDL desktop icon or run
;     IDL from the Start->Programs menu; on Unix, at the Unix command
;     prompt, type
;         idlde
;  2) From the File item on the menu bar, choose Open Project
;     and then select the file subset_sage2_v620.prj
;     Note:  If your preferences (File->Preferences) are set with a check
;     mark next to Change Directory on Open, the directory containing
;     this file will now be the IDL default directory.
;  2) From the Project item on the menu bar, choose Build and then choose
;     Run.  The program will be running in the mode where the SAGE II files
;     to be processed can be selected from a dialog box.
;
;  To run with a list file, Build the program as indicated above,
;  then, at the IDL command line (indicated with the IDL> prompt), type
;      subset_sage2_v620,LISTFILE="your_listfile"
;  where your_listfile is the name (including path if not in the default
;  IDL directory) of the file containing the names of SAGE II species
;  files to be processed.
;
;  The program executes as follows:
;  1)  If the LISTFILE keyword is not specified, a file selection dialog
;      box appears which may be used to choose one or more SAGE II species
;      files for processing.
;  2)  A selection box is displayed for entering latitude and longitude
;      bounds for the subset.  By default, the entire globe is selected
;      (latitudes -90 to 90, longitudes -180 to 180).
;  3)  A parameter selection box is displayed.  One or more parameters
;      may be selected for inclusion in the subset file.
;
;  An hourglass cursor is displayed while the files are processed.
;  A message box appears when processing is complete.
;
;  HISTORY:
;  Original version:  April 2004

;  if running via the IDL Virtual Machine, set the default directory to
;  the directory where this routine is located (otherwise it will default
;  to the RSI IDL binary directory)

;  if not running the IDL VM, the default directory will be determined
;  by the user's IDL preferences or current setting

if LMGR(/VM) then begin
  info = ROUTINE_INFO('subset_sage2_v620',/source)
  vmdir = FILE_DIRNAME(info.path)
  cd,vmdir
endif

num_params = 32

;  get lat/lon region of interest

north_lat = 90
south_lat = -90
west_lon = -180
east_lon = 180
get_latlon_box,north_lat,south_lat,west_lon,east_lon

;  get parameters of interest

param_list = STRARR(num_params)
OPENR,plun,'sage2_params.txt',/GET_LUN
READF,plun,param_list
FREE_LUN,plun

select_tool,param_list,selected_params,TITLE="Select parameters to include in subset"

pnames = STRARR(num_params)
for i=0,31 do pnames[i] = STRMID(param_list[i],0,strpos(param_list[i],':'))

if KEYWORD_SET(LISTFILE) then begin
  ;  determine the number of files in the list file
  ;  (done in this manner rather than using the IDL FILE_LINES function
  ;  to accommodate older versions of IDL)
  OPENR,listlun,listfile,/GET_LUN
  num_files = 0
  while not EOF(listlun) do begin
    file = ''
    readf,listlun,file
    num_files = num_files + 1
  endwhile
  FREE_LUN,listlun

  ;  now read the file names into the files array
  OPENR,listlun,listfile,/GET_LUN
  files = STRARR(num_files)
  readf,listlun,files
  FREE_LUN,listlun

endif else begin
  ;  if list file is not specified, assume this is to be done for a single file and
  ;  get file name
  files = DIALOG_PICKFILE(TITLE="Select SAGE II Species File",$
            FILTER="*SPEC*",/MULTIPLE_FILES)
  num_files = n_elements(files)

endelse

;  for each file to be processed, get records that match lat/lon spec
;  and extract desired parameters

WIDGET_CONTROL,/HOURGLASS

for i=0,num_files-1 do begin
  specfile = files[i]
  specloc = STRPOS(specfile,"SPEC")
  ;  check for selection of a species file--if not a species file, note that and go to next file
  if (specloc lt 0) then begin
    printf,dlun,specfile," is not a valid SAGE II SPEC file--skipping the file"
    goto, endfile
  endif

  ;  determine the name of the corresponding index file
  indexfile = STRMID(specfile,0,specloc) + "INDEX" + STRMID(specfile,specloc+4,STRLEN(specfile)-1)

  ;  determine the name for the output file
  subsetfile = specfile + "_subset.dat"

  ;  print subset information to output file

  openw,dlun,subsetfile,/GET_LUN
  printf,dlun,"Subset of File ",specfile
  printf,dlun,"Subset criteria: "
  printf,dlun,"Latitude:  ",south_lat, " - ",north_lat
  printf,dlun,"Longitude: ",west_lon, " - ",east_lon
  printf,dlun,"Included parameters: "

  for p = 0,n_elements(selected_params) -1 do printf,dlun,"  ",pnames[selected_params[p]]

  ;  read the contents of the index file
  readstructs,indexfile,index

  ;  find profiles whose lat/lon values meet user criteria
  ;  but don't care if the user didn't specify a lat/lon subset

  if (south_lat ne -90) or (north_lat ne 90) or (west_lon ne -180) or (east_lon ne 180) then begin

    ;  find profiles with matching latitudes

    goodlats = WHERE ((index.lat ge south_lat) and (index.lat le north_lat),latcount)
    if (latcount le 0) then begin
      printf,dlun,"No latitudes found within specified range of ",south_lat,north_lat
      goto, endfile
    endif
    latmask = BYTARR(index.num_prof) ; zeroed out by default
    latmask[goodlats] = 1

    ;  find profiles with matching longitudes, taking into account
    ;  a longitude specification that crosses the Internationl Date Line

    if (east_lon lt west_lon) then begin  ; longitude range crosses the dateline
      goodlons = WHERE ((((index.lon ge east_lon) and (index.lon le 180)) or $
                        ((index.lon ge -180) and (index.lon le west_lon))),loncount)
    endif else begin
      goodlons = WHERE (((index.lon ge west_lon) and (index.lon le east_lon)),loncount)
    endelse

    if (loncount le 0) then begin
      printf,dlun,"No longitudes found within specified range of ",west_lon,east_lon
      goto, endfile
    endif
    lonmask = BYTARR(index.num_prof) ; zeroed out by default
    lonmask[goodlons]  = 1

    ;  determine profiles that match both lat and lon criteria
    recmask = latmask and lonmask

  endif else begin
    ;  a lat/lon subset was not specified so want all profiles
    recmask = BYTARR(index.num_prof)
    recmask[*] = 1
  endelse

  ;  get profile numbers for good profiles
  goodrec = WHERE(recmask eq 1,reccount)
  if (reccount le 0) then begin
    printf,dlun,"No profiles have both latitude and longitude within the specified range"
    goto, endfile
  endif

  ;  print info from index file that applies to the entire file (first 22 items)
  ;  as output file header

  index_tagnames = tag_names(index)
  printf,dlun,""
  printf,dlun,"FILE INFORMATION"
  printf,dlun,"----------------"
  for k = 0,21 do begin
    type = size(index.(k),/type)
    if type eq 1 then begin
      ;  if it is a byte array, print the string
      printf,dlun,index_tagnames[k],'     ',string(index.(k))
    endif else if n_elements(index.(k) eq 1) then begin
      ;  if it has just one element, print all on one line
      printf,dlun,index_tagnames[k],index.(k)
    endif else begin
      ;  for multiple elements, print the title on a separate line
      printf,dlun,index_tagnames[k]
      printf,dlun,index.(k)
    endelse
  endfor

  printf,dlun,""

  for j = 0,reccount-1 do begin
    printf,dlun,"------------"
    printf,dlun,"Profile number: ",goodrec[j]+1
    ;  print profile specific info from index file
    for k = 22,31 do begin
      printf,dlun,index_tagnames[k],index.(k)(goodrec[j])
    endfor
    printf,dlun,""

    ;  get the profile
    readstructs,specfile,goodrec[j],onerec

    ;  print out data for selected parameters

    spec_tagnames = TAG_NAMES(onerec)
    for k=0,N_ELEMENTS(selected_params)-1 do begin
      printf,dlun,spec_tagnames[selected_params[k]]
      printf,dlun,onerec.(selected_params[k])
    endfor ; loop over selected parameters
  endfor  ; loop over selected profiles

  endfile: if (n_elements(dlun) gt 0) then FREE_LUN,dlun

endfor  ; loop over selected files

response = DIALOG_MESSAGE("Processing complete",/INFO)

end